"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _slicedToArray = function () { function sliceIterator(arr, i) { var _arr = []; var _n = true; var _d = false; var _e = undefined; try { for (var _i = arr[Symbol.iterator](), _s; !(_n = (_s = _i.next()).done); _n = true) { _arr.push(_s.value); if (i && _arr.length === i) break; } } catch (err) { _d = true; _e = err; } finally { try { if (!_n && _i["return"]) _i["return"](); } finally { if (_d) throw _e; } } return _arr; } return function (arr, i) { if (Array.isArray(arr)) { return arr; } else if (Symbol.iterator in Object(arr)) { return sliceIterator(arr, i); } else { throw new TypeError("Invalid attempt to destructure non-iterable instance"); } }; }();

var _createClass = function () { function defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } } return function (Constructor, protoProps, staticProps) { if (protoProps) defineProperties(Constructor.prototype, protoProps); if (staticProps) defineProperties(Constructor, staticProps); return Constructor; }; }();

var _typeof = typeof Symbol === "function" && typeof Symbol.iterator === "symbol" ? function (obj) { return typeof obj; } : function (obj) { return obj && typeof Symbol === "function" && obj.constructor === Symbol && obj !== Symbol.prototype ? "symbol" : typeof obj; }; /**
                                                                                                                                                                                                                                                                               * Wrapper for Kurento's PeerConnection, which in turn wraps and abstracts the actual WebRTC
                                                                                                                                                                                                                                                                               * interfaces.
                                                                                                                                                                                                                                                                               */

var _debug = require("debug");

var _debug2 = _interopRequireDefault(_debug);

var _spClient = require("sp-client");

var _spClient2 = _interopRequireDefault(_spClient);

var _wolfy87Eventemitter = require("wolfy87-eventemitter");

var _wolfy87Eventemitter2 = _interopRequireDefault(_wolfy87Eventemitter);

var _kurentoUtils = require("kurento-utils");

var _kurentoUtils2 = _interopRequireDefault(_kurentoUtils);

var _url = require("url");

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _defineProperty(obj, key, value) { if (key in obj) { Object.defineProperty(obj, key, { value: value, enumerable: true, configurable: true, writable: true }); } else { obj[key] = value; } return obj; }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

function _possibleConstructorReturn(self, call) { if (!self) { throw new ReferenceError("this hasn't been initialised - super() hasn't been called"); } return call && (typeof call === "object" || typeof call === "function") ? call : self; }

function _inherits(subClass, superClass) { if (typeof superClass !== "function" && superClass !== null) { throw new TypeError("Super expression must either be null or a function, not " + typeof superClass); } subClass.prototype = Object.create(superClass && superClass.prototype, { constructor: { value: subClass, enumerable: false, writable: true, configurable: true } }); if (superClass) Object.setPrototypeOf ? Object.setPrototypeOf(subClass, superClass) : subClass.__proto__ = superClass; }

var log = (0, _debug2.default)("sp:sp-peer-connection");
var kurentoLog = (0, _debug2.default)("sp:sp-peer-kurento");

// This is the only way we get any control of Kurento's logger... see also index.html
if (_typeof(window.Logger) === "object") {
  Object.keys(console).forEach(function (key) {
    window.Logger[key] = kurentoLog;
  });
}

if (typeof RTCPeerConnection === "undefined") {
  if (typeof webkitRTCPeerConnection !== "undefined") {
    window.RTCPeerConnection = window.webkitRTCPeerConnection;
  } else if (typeof mozRTCPeerConnection !== "undefined") {
    window.RTCPeerConnection = window.mozRTCPeerConnection;
  } else {
    _spClient2.default.error("No RTCPeerConnection implementation found");
  }
}

var PEER_CONNECTION_EVENTS = ["addstream", "icecandidate", "iceconnectionstatechange", "icegatheringstatechange", "negotiationneeded", "removestream", "signalingstatechange", "datachannel", "close", "error", "message", "open", "tonechange", "identityresult", "idpassertionerror", "idpvalidationerror", "peeridentity", "isolationchange"];

var SPPeerConnection = function (_EE) {
  _inherits(SPPeerConnection, _EE);

  function SPPeerConnection(_ref) {
    var targetUserId = _ref.targetUserId,
        stream = _ref.stream;

    _classCallCheck(this, SPPeerConnection);

    var _this = _possibleConstructorReturn(this, (SPPeerConnection.__proto__ || Object.getPrototypeOf(SPPeerConnection)).call(this));

    _this.isShutDown = false;
    _this.targetUserId = targetUserId;
    _this.videoStream = stream;
    _this.localIceCandidates = [];
    _this.remoteIceCandidatesChecked = [];
    _this.processedOffer = false;
    _this.kurentoIsSetup = false;
    _this.options = {
      mediaConstraints: {
        audio: true,
        video: {
          width: 1280,
          height: 720
        }
      },
      onicecandidate: _this.onIceCandidate.bind(_this),
      configuration: {
        iceTransportPolicy: "relay"
      }
    };

    _this.streamPromise = new Promise(function (resolve, reject) {
      _this._streamResolve = resolve;
      _this._streamReject = reject;
    });

    _this.isPrimaryUser = _spClient2.default.user.id < targetUserId;
    if (_this.isPrimaryUser) {
      _this.createOffer();
    } else {
      _this.waitForOffer();
    }

    // Set a timeout that retries everything if nothing's worked. And do it randomly, so you don't
    // get two peers timing out and retrying at the same time.
    _this.timeoutMs = 5000 + Math.floor(Math.random() * 5000);
    _this.resetTimeout();
    _this.shouldTimeout = true;
    return _this;
  }

  /**
   * Reset the auto-retry timeout
   */


  _createClass(SPPeerConnection, [{
    key: "resetTimeout",
    value: function resetTimeout() {
      var _this2 = this;

      if (!this.shouldTimeout) {
        return;
      }
      if (this.timeoutHandle) {
        clearTimeout(this.timeoutHandle);
      }
      this.timeoutHandle = setTimeout(function () {
        log("Hit timeout of " + _this2.timeoutMs + "ms, aborting connection.");
        _this2.shutdown();
      }, this.timeoutMs);
    }

    /**
     * Stop the auto-retry timeout
     */

  }, {
    key: "stopTimeout",
    value: function stopTimeout() {
      if (this.timeoutHandle) {
        clearTimeout(this.timeoutHandle);
      }
      this.shouldTimeout = false;
    }

    /**
     * Stop everything about this connection and clean everything up.
     */

  }, {
    key: "shutdown",
    value: function shutdown() {
      if (this.isShutDown === true) {
        return;
      }
      this.stopTimeout();
      this.webRtcPeer && this.webRtcPeer.dispose();
      this.isShutDown = true;
      this.peerHandle && this.peerHandle.stop();
      this.emit("disconnected");
      if (this.peer) {
        _spClient2.default.peerconnections.delete(this.peer.id);
      }
    }
  }, {
    key: "_generateKurentoPeer",
    value: function _generateKurentoPeer() {
      var _this3 = this;

      return new Promise(function (resolve, reject) {
        log("Creating Kurento peer");
        _this3.webRtcPeer = _kurentoUtils2.default.WebRtcPeer.WebRtcPeerSendrecv(_this3.options, function (error) {
          _this3.webRtcPeer.peerConnection.addEventListener("iceconnectionstatechange", _this3.handleConnectionChange.bind(_this3));
          if (error) {
            return reject(error);
          }
          _this3.kurentoIsSetup = true;
          resolve();
        });
      });
    }
  }, {
    key: "handleConnectionChange",
    value: function handleConnectionChange(e) {
      var state = e.currentTarget.iceConnectionState;
      log("ICE Connection State: " + state);
      if (state === "connected") {
        this.stopTimeout();
      }
      if (state === "completed") {
        // Cool, we're connected, no reason to keep that peerconnection around anymore
        this.peerHandle.stop();
        // Commented for now; needs work. Things get deleted at the wrong time and cycle.
        // if (this.isPrimaryUser) {
        //   SP.peerconnections.delete(this.peer.id).catch(log);
        // }
      }
      if (state === "failed" || state === "disconnected" || state === "closed") {
        this.shutdown();
      }
    }

    /**
     * We're the smaller one. Create an API object for it.
     */

  }, {
    key: "createOffer",
    value: function createOffer() {
      var _this4 = this;

      log("Creating peerconnection");
      return _spClient2.default.peerconnections.create({
        userId: _spClient2.default.user.id,
        targetUserId: this.targetUserId
      }).then(function (peer) {
        _this4.resetTimeout();
        _this4.peer = peer;
        _this4.peerHandle = _spClient2.default.peerconnections.watch({ id: peer.id });
        _this4.peerHandle.on("data", _this4.peerUpdate.bind(_this4));
        return _this4.setupTurnUrls();
      }).then(function () {
        _this4.resetTimeout();
        log("Generating offer");
        return new Promise(function (resolve, reject) {
          _this4.localIceCandidates = [];
          _this4.webRtcPeer.generateOffer(function (err, offer) {
            if (err) {
              return reject(err);
            }
            resolve(offer);
          });
        });
      }).then(function (offer) {
        _this4.resetTimeout();
        return _spClient2.default.peerconnections.update(_this4.peer.id, {
          sdpOffer: offer,
          userIceCandidates: _this4.localIceCandidates
        });
      }).then(function (peer) {
        _this4.resetTimeout();
        log("Created peerconnection " + peer.id);
        return new Promise(function (resolve, reject) {
          _this4.peerHandle = _spClient2.default.peerconnections.watch({ id: peer.id }).catch(log).on("data", function (_ref2) {
            var _ref3 = _slicedToArray(_ref2, 1),
                doc = _ref3[0];

            if (!doc) {
              // Hmm, our candidate is just gone. Annoying and mysterious. Well, start the process
              // over again.
              log("Peer connection " + peer.id + " was deleted, retrying.");
              _this4.peerHandle.stop();
              // FIXME: we now just have a promise sitting around? does that matter?
              _this4.createOffer();
            }
            if (doc.sdpAnswer) {
              _this4.peerHandle.stop();
              resolve(doc);
            }
          });
        });
      }).then(function (peer) {
        _this4.resetTimeout();
        log("Got peer " + peer.id);
        _this4.peerHandle = _spClient2.default.peerconnections.watch({ id: peer.id }).on("data", _this4.peerUpdate.bind(_this4));
        return _this4.peerHandle;
      }).then(function (peer) {
        _this4.resetTimeout();
        return _spClient2.default.peerconnections.update(_this4.peer.id, {
          userIceCandidates: _this4.localIceCandidates
        });
      }).catch(function (err) {
        log(err);
      });
    }

    /**
     * We're the bigger one. Wait for an API object from the target user.
     * @return {[type]} [description]
     */

  }, {
    key: "waitForOffer",
    value: function waitForOffer() {
      var _this5 = this;

      log("Waiting for offer from " + this.targetUserId);
      this.peerHandle = _spClient2.default.peerconnections.watch({
        userId: this.targetUserId,
        targetUserId: _spClient2.default.user.id
      });
      return this.peerHandle.then(function (docs) {
        // Hack: for now, delete all the other requests that were here when we got here...
        // This will change once we have a better API notion of what it means to have one user
        // with multiple browsers open
        docs.forEach(function (doc) {
          _spClient2.default.peerconnections.delete(doc.id).then(function () {
            log("deleted stale icecandidate " + doc.id);
          }).catch(log);
        });
        return new Promise(function (resolve, reject) {
          _this5.peerHandle.on("newDoc", function (data) {
            _this5.peerHandle.stop();
            resolve(data);
          });
        });
      }).then(function (peer) {
        _this5.resetTimeout();
        _this5.peer = peer;
        log("Got peer " + peer.id);
        _this5.peerHandle = _spClient2.default.peerconnections.watch({ id: _this5.peer.id }).on("data", _this5.peerUpdate.bind(_this5));
        return _this5.setupTurnUrls();
      }).catch(log);
    }

    /**
     * Wait on this.peerHandle until we've been assigned turnUrls from the server.
     */

  }, {
    key: "setupTurnUrls",
    value: function setupTurnUrls() {
      var _this6 = this;

      return new Promise(function (resolve, reject) {
        if (_this6.peer && _this6.peer.turnUrls && _this6.peer.turnUrls.length > 1) {
          return resolve(_this6.peer.turnUrls);
        }
        var onData = function onData(_ref4) {
          var _ref5 = _slicedToArray(_ref4, 1),
              doc = _ref5[0];

          if (!doc) {
            return _this6.peerHandle.off("data", onData);
          }
          if (!doc.turnUrls || doc.turnUrls.length < 1) {
            return;
          }
          _this6.peerHandle.off("data", onData);
          resolve(doc.turnUrls);
        };
        _this6.peerHandle.on("data", onData);
      }).then(function (turnUrls) {
        _this6.options.configuration.iceServers = turnUrls.map(function (turnUrl) {
          var _parseUrl = (0, _url.parse)(turnUrl),
              protocol = _parseUrl.protocol,
              auth = _parseUrl.auth,
              host = _parseUrl.host;

          var _auth$split = auth.split(":"),
              _auth$split2 = _slicedToArray(_auth$split, 2),
              username = _auth$split2[0],
              credential = _auth$split2[1];

          return {
            url: "" + protocol + host,
            credential: credential,
            username: username
          };
        });
        return _this6._generateKurentoPeer();
      }).then(function () {
        _this6.peerUpdate([_this6.peer]);
      });
    }

    /**
     * Code to run every time our salient PeerConnection changes.
     */

  }, {
    key: "peerUpdate",
    value: function peerUpdate(_ref6) {
      var _this7 = this;

      var _ref7 = _slicedToArray(_ref6, 1),
          doc = _ref7[0];

      this.resetTimeout();
      if (!doc) {
        return this.shutdown();
      }
      this.peer = doc;
      // If we're waiting on their answer and haven't processed it yet...
      if (this.kurentoIsSetup && !this.processedOffer) {
        if (this.isPrimaryUser && this.peer.sdpAnswer) {
          this.processedOffer = true;
          return this.webRtcPeer.processAnswer(this.peer.sdpAnswer, function (err, answer) {
            var remoteStream = _this7.webRtcPeer.getRemoteStream();
            log("Remote stream", remoteStream);
            _this7._streamResolve(remoteStream);
          });
        } else if (!this.isPrimaryUser && this.peer.sdpOffer) {
          this.processedOffer = true;
          return new Promise(function (resolve, reject) {
            _this7.webRtcPeer.processOffer(_this7.peer.sdpOffer, function (err, answer) {
              if (err) {
                reject(err);
              }
              _this7._streamResolve(_this7.webRtcPeer.getRemoteStream());
              resolve(answer);
            });
          }).then(function (answer) {
            _this7.resetTimeout();
            return _spClient2.default.peerconnections.update(_this7.peer.id, {
              sdpAnswer: answer,
              targetUserIceCandidates: _this7.localIceCandidates
            });
          }).catch(function (err) {
            log(err);
            _this7.shutdown();
          });
        }
      }
      if (this.kurentoIsSetup) {
        var field = this.isPrimaryUser ? "targetUserIceCandidates" : "userIceCandidates";
        doc[field] && doc[field].forEach(this.onRemoteIceCandidate.bind(this));
      }
    }
  }, {
    key: "onIceCandidate",
    value: function onIceCandidate(candidate) {
      log("ice", candidate);
      this.localIceCandidates.push(candidate);
      // If we're not connected yet, return.
      if (!this.peer) {
        return;
      }
      var field = this.isPrimaryUser ? "userIceCandidates" : "targetUserIceCandidates";
      _spClient2.default.peerconnections.update(this.peer.id, _defineProperty({}, field, this.localIceCandidates)).catch(log);
    }
  }, {
    key: "onRemoteIceCandidate",
    value: function onRemoteIceCandidate(candidate) {
      var str = JSON.stringify(candidate);
      if (this.remoteIceCandidatesChecked.indexOf(str) !== -1) {
        // Already checked.
        return;
      }
      log("Evaluating remote ICE candidate", candidate);
      this.remoteIceCandidatesChecked.push(candidate);
      this.webRtcPeer.addIceCandidate(candidate);
    }
  }, {
    key: "getStream",
    value: function getStream() {
      return this.streamPromise;
    }
  }]);

  return SPPeerConnection;
}(_wolfy87Eventemitter2.default);

exports.default = SPPeerConnection;